import functools
import itertools
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional

import colorama

from semgrep.constants import BREAK_LINE
from semgrep.constants import BREAK_LINE_CHAR
from semgrep.constants import BREAK_LINE_WIDTH
from semgrep.constants import CLI_RULE_ID
from semgrep.constants import ELLIPSIS_STRING
from semgrep.constants import MAX_CHARS_FLAG_NAME
from semgrep.constants import MAX_LINES_FLAG_NAME
from semgrep.formatter.base import BaseFormatter
from semgrep.rule_match import RuleMatch
from semgrep.util import format_bytes
from semgrep.util import truncate


class TextFormatter(BaseFormatter):
    @staticmethod
    def _color_line(
        line: str,
        line_number: int,
        start_line: int,
        start_col: int,
        end_line: int,
        end_col: int,
    ) -> str:
        start_color = 0 if line_number > start_line else start_col
        # column offset
        start_color = max(start_color - 1, 0)
        end_color = end_col if line_number >= end_line else len(line) + 1 + 1
        end_color = max(end_color - 1, 0)
        line = (
            line[:start_color]
            + colorama.Style.BRIGHT
            + line[start_color : end_color + 1]  # want the color to include the end_col
            + colorama.Style.RESET_ALL
            + line[end_color + 1 :]
        )
        return line

    @staticmethod
    def _finding_to_line(
        rule_match: RuleMatch,
        color_output: bool,
        per_finding_max_lines_limit: Optional[int],
        per_line_max_chars_limit: Optional[int],
        show_separator: bool,
    ) -> Iterator[str]:
        path = rule_match.path
        start_line = rule_match.start.get("line")
        end_line = rule_match.end.get("line")
        start_col = rule_match.start.get("col")
        end_col = rule_match.end.get("col")
        trimmed = 0
        stripped = False
        if path:
            lines = rule_match.extra.get("fixed_lines") or rule_match.lines
            if per_finding_max_lines_limit:
                trimmed = len(lines) - per_finding_max_lines_limit
                lines = lines[:per_finding_max_lines_limit]

            for i, line in enumerate(lines):
                line = line.rstrip()
                line_number = ""
                if start_line:
                    if color_output:
                        line = TextFormatter._color_line(
                            line, start_line + i, start_line, start_col, end_line, end_col  # type: ignore
                        )
                        line_number = f"{colorama.Fore.GREEN}{start_line + i}{colorama.Style.RESET_ALL}"
                    else:
                        line_number = f"{start_line + i}"

                    if (
                        per_line_max_chars_limit
                        and len(line) > per_line_max_chars_limit
                    ):
                        stripped = True
                        is_first_line = i == 0
                        if is_first_line:
                            line = (
                                line[
                                    start_col - 1 : start_col - 1 + per_line_max_chars_limit  # type: ignore
                                ]
                                + ELLIPSIS_STRING
                            )
                            if start_col > 1:  # type: ignore
                                line = ELLIPSIS_STRING + line
                        else:
                            line = line[:per_line_max_chars_limit] + ELLIPSIS_STRING
                        # while stripping a string, the ANSI code for resetting color might also get stripped.
                        line = line + colorama.Style.RESET_ALL

                yield f"{line_number}:{line}" if line_number else f"{line}"

            if stripped:
                yield f"[Shortened a long line from output, adjust with {MAX_CHARS_FLAG_NAME}]"
            trimmed_str = (
                f" [hid {trimmed} additional lines, adjust with {MAX_LINES_FLAG_NAME}] "
            )
            if per_finding_max_lines_limit != 1:
                if trimmed > 0:
                    yield trimmed_str.center(BREAK_LINE_WIDTH, BREAK_LINE_CHAR)
                elif show_separator:
                    yield BREAK_LINE

    @staticmethod
    def _build_text_timing_output(
        time_data: Dict[str, Any],
        color_output: bool,
    ) -> Iterator[str]:
        items_to_show = 5
        col_lim = 70
        RESET_COLOR = colorama.Style.RESET_ALL if color_output else ""
        GREEN_COLOR = colorama.Fore.GREEN if color_output else ""
        YELLOW_COLOR = colorama.Fore.YELLOW if color_output else ""

        rule_parsing_time = sum(
            parse_time for parse_time in time_data["rule_parse_info"]
        )
        rule_timings = {
            rule["id"]: functools.reduce(
                lambda x, y: (x[0] + y[0], x[1] + y[1]),
                (
                    (t["run_times"][i] - t["parse_times"][i], t["match_times"][i])
                    for t in time_data["targets"]
                ),
                (time_data["rule_parse_info"][i], 0.0),
            )
            for i, rule in enumerate(time_data["rules"])
        }
        file_parsing_time = sum(
            sum(target["parse_times"]) for target in time_data["targets"]
        )
        file_timings = {
            target["path"]: float(sum(target["run_times"]))
            for target in time_data["targets"]
        }

        all_total_time = sum(i for i in file_timings.values()) + rule_parsing_time
        total_matching_time = sum(i[1] for i in rule_timings.values())

        # Output information
        yield f"\nSemgrep-core timing summary:"
        yield f"Total CPU time: {all_total_time:.4f}  File parse time: {file_parsing_time:.4f}" f"  Rule parse time: {rule_parsing_time:.4f}  Match time: {total_matching_time:.4f}"

        yield f"Slowest {items_to_show}/{len(file_timings)} files"
        slowest_file_times = sorted(
            file_timings.items(), key=lambda x: x[1], reverse=True
        )[:items_to_show]
        for file_name, parse_time in slowest_file_times:
            num_bytes = f"({format_bytes(Path(file_name).resolve().stat().st_size)}):"
            file_name = truncate(file_name, col_lim)
            yield f"{GREEN_COLOR}{file_name:<70}{RESET_COLOR} {num_bytes:<9}{parse_time:.4f}"

        yield f"Slowest {items_to_show} rules to run (excluding parse time)"
        slowest_rule_times = sorted(
            rule_timings.items(), key=lambda x: float(x[1][0]), reverse=True
        )[:items_to_show]
        for rule_id, (total_time, match_time) in slowest_rule_times:
            rule_id = truncate(rule_id, col_lim) + ":"
            yield f"{YELLOW_COLOR}{rule_id:<71}{RESET_COLOR} run time {total_time:.4f}  match time {match_time:.4f}"

    @staticmethod
    def _build_text_output(
        rule_matches: List[RuleMatch],
        color_output: bool,
        per_finding_max_lines_limit: Optional[int],
        per_line_max_chars_limit: Optional[int],
    ) -> Iterator[str]:
        RESET_COLOR = colorama.Style.RESET_ALL if color_output else ""
        GREEN_COLOR = colorama.Fore.GREEN if color_output else ""
        YELLOW_COLOR = colorama.Fore.YELLOW if color_output else ""
        RED_COLOR = colorama.Fore.RED if color_output else ""
        BLUE_COLOR = colorama.Fore.BLUE if color_output else ""

        last_file = None
        last_message = None
        sorted_rule_matches = sorted(rule_matches, key=lambda r: (r.path, r.id))
        for rule_index, rule_match in enumerate(sorted_rule_matches):

            current_file = rule_match.path
            check_id = rule_match.id
            message = rule_match.message
            severity = rule_match.severity.lower()
            fix = rule_match.fix
            if last_file is None or last_file != current_file:
                if last_file is not None:
                    yield ""
                yield f"{GREEN_COLOR}{current_file}{RESET_COLOR}"
                last_message = None
            # don't display the rule line if the check is empty
            if (
                check_id
                and check_id != CLI_RULE_ID
                and (last_message is None or last_message != message)
            ):
                severity_prepend = ""
                if severity:
                    if severity == "error":
                        severity_prepend = f"{RED_COLOR}severity:{severity} "
                    elif severity == "warning":
                        severity_prepend = f"{YELLOW_COLOR}severity:{severity} "
                    else:
                        severity_prepend = f"severity:{severity} "
                yield f"{severity_prepend}{YELLOW_COLOR}rule:{check_id}: {message}{RESET_COLOR}"

            last_file = current_file
            last_message = message
            next_rule_match = (
                sorted_rule_matches[rule_index + 1]
                if rule_index != len(sorted_rule_matches) - 1
                else None
            )

            if fix:
                yield f"{BLUE_COLOR}autofix:{RESET_COLOR} {fix}"
            elif rule_match.fix_regex:
                fix_regex = rule_match.fix_regex
                yield f"{BLUE_COLOR}autofix:{RESET_COLOR} s/{fix_regex.get('regex')}/{fix_regex.get('replacement')}/{fix_regex.get('count', 'g')}"

            is_same_file = (
                next_rule_match.path == rule_match.path if next_rule_match else False
            )
            yield from TextFormatter._finding_to_line(
                rule_match,
                color_output,
                per_finding_max_lines_limit,
                per_line_max_chars_limit,
                is_same_file,
            )

    def output(self) -> str:
        output = self._build_text_output(
            self.rule_matches,
            self.extra["color_output"],
            self.extra["per_finding_max_lines_limit"],
            self.extra["per_line_max_chars_limit"],
        )

        timing_output = (
            self._build_text_timing_output(
                self.extra["time"],
                self.extra["color_output"],
            )
            if "time" in self.extra
            else iter([])
        )

        return "\n".join(itertools.chain(output, timing_output))



from typing import Dict
from typing import List

from semgrep.error import _UnknownExtensionError
from semgrep.error import _UnknownLanguageError
from semgrep.semgrep_types import FileExtension
from semgrep.semgrep_types import Language
from semgrep.semgrep_types import Language_util


# coupling: if you add a constant here, modify also ALL_EXTENSIONS below
# and you probably also need to update semgrep_types.py Language
# and Language_util classes.
# You may also have to regenerate some test snapshots with
# pipenv run pytest tests/e2e/test_rule_parser.py --snapshot-update
PYTHON_EXTENSIONS = [FileExtension(".py"), FileExtension(".pyi")]
JAVASCRIPT_EXTENSIONS = [FileExtension(".js"), FileExtension(".jsx")]
TYPESCRIPT_EXTENSIONS = [FileExtension(".ts"), FileExtension(".tsx")]
JAVA_EXTENSIONS = [FileExtension(".java")]
C_EXTENSIONS = [FileExtension(".c")]
GO_EXTENSIONS = [FileExtension(".go")]
RUBY_EXTENSIONS = [FileExtension(".rb")]
PHP_EXTENSIONS = [FileExtension(".php")]
LUA_EXTENSIONS = [FileExtension(".lua")]
CSHARP_EXTENSIONS = [FileExtension(".cs")]
RUST_EXTENSIONS = [FileExtension(".rs")]
KOTLIN_EXTENSIONS = [FileExtension(".kt"), FileExtension(".kts"), FileExtension(".ktm")]
YAML_EXTENSIONS = [FileExtension(".yaml"), FileExtension(".yml")]
ML_EXTENSIONS = [FileExtension(".mli"), FileExtension(".ml")]
JSON_EXTENSIONS = [FileExtension(".json")]
SCALA_EXTENSIONS = [FileExtension(".scala")]
VUE_EXTENSIONS = [FileExtension(".vue")]
HTML_EXTENSIONS = [FileExtension(".html"), FileExtension(".html")]

# This is used to determine the set of files with known extensions,
# i.e. those for which we have a proper parser.
ALL_EXTENSIONS = (
    PYTHON_EXTENSIONS
    + JAVASCRIPT_EXTENSIONS
    + TYPESCRIPT_EXTENSIONS
    + JAVA_EXTENSIONS
    + C_EXTENSIONS
    + GO_EXTENSIONS
    + RUBY_EXTENSIONS
    + ML_EXTENSIONS
    + JSON_EXTENSIONS
    + RUST_EXTENSIONS
    + KOTLIN_EXTENSIONS
    + YAML_EXTENSIONS
    + SCALA_EXTENSIONS
    + VUE_EXTENSIONS
    + HTML_EXTENSIONS
)

# This is used to select the files suitable for spacegrep, which is
# all of them. It is spacegrep itself that will detect and ignore binary
# files.
GENERIC_EXTENSIONS = [FileExtension("")]

_LANGS_TO_EXTS: Dict[Language, List[FileExtension]] = {
    Language.PYTHON: PYTHON_EXTENSIONS,
    Language.PYTHON2: PYTHON_EXTENSIONS,
    Language.PYTHON3: PYTHON_EXTENSIONS,
    Language.JAVASCRIPT: JAVASCRIPT_EXTENSIONS,
    Language.TYPESCRIPT: TYPESCRIPT_EXTENSIONS,
    Language.JAVA: JAVA_EXTENSIONS,
    Language.C: C_EXTENSIONS,
    Language.GO: GO_EXTENSIONS,
    Language.ML: ML_EXTENSIONS,
    Language.RUBY: RUBY_EXTENSIONS,
    Language.PHP: PHP_EXTENSIONS,
    Language.JSON: JSON_EXTENSIONS,
    Language.LUA: LUA_EXTENSIONS,
    Language.CSHARP: CSHARP_EXTENSIONS,
    Language.RUST: RUST_EXTENSIONS,
    Language.KOTLIN: KOTLIN_EXTENSIONS,
    Language.YAML: YAML_EXTENSIONS,
    Language.REGEX: GENERIC_EXTENSIONS,
    Language.GENERIC: GENERIC_EXTENSIONS,
    Language.SCALA: SCALA_EXTENSIONS,
    Language.VUE: VUE_EXTENSIONS,
    Language.HTML: HTML_EXTENSIONS,
}


def all_supported_languages() -> List[str]:
    """
    We want the list of languages to be deterministic, so sort it
    """
    return Language_util.all_language_strs()


# create a dictionary for fast lookup and reverse lookup
_EXTS_TO_LANG: Dict[FileExtension, Language] = {}
for language in _LANGS_TO_EXTS.keys():
    for extension in _LANGS_TO_EXTS[language]:
        # When there are multiple languages for an extension,
        # take the first one
        if not extension in _EXTS_TO_LANG:
            _EXTS_TO_LANG[extension] = language


def ext_to_lang(ext: FileExtension) -> Language:
    langs = _EXTS_TO_LANG.get(ext)
    if langs is None:
        raise _UnknownExtensionError(f"Unsupported extension: {ext}")
    return langs


def lang_to_exts(language: Language) -> List[FileExtension]:
    """
    Convert language to expected file extensions

    If language is not a supported semgrep language then
    raises _UnknownLanguageError
    """
    extensions = _LANGS_TO_EXTS.get(language)
    if extensions is None:
        raise _UnknownLanguageError(f"Unsupported Language: {language}")
    return extensions



import abc
from typing import Any
from typing import Dict
from typing import FrozenSet
from typing import List

from semgrep.error import SemgrepError
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch


class BaseFormatter(abc.ABC):
    def __init__(
        self,
        rules: FrozenSet[Rule],
        rule_matches: List[RuleMatch],
        semgrep_structured_errors: List[SemgrepError],
        extra: Dict[str, Any],
    ) -> None:
        self.rules = rules
        self.rule_matches = rule_matches
        self.semgrep_structured_errors = semgrep_structured_errors
        self.extra = extra

    @abc.abstractmethod
    def output(self) -> str:
        raise NotImplementedError



from typing import List

from semgrep.formatter.base import BaseFormatter
from semgrep.rule_match import RuleMatch


class VimFormatter(BaseFormatter):
    @staticmethod
    def _get_parts(rule_match: RuleMatch) -> List[str]:
        severity = {
            "INFO": "I",
            "WARNING": "W",
            "ERROR": "E",
        }
        return [
            str(rule_match.path),
            str(rule_match.start["line"]),
            str(rule_match.start["col"]),
            severity[rule_match.severity],
            rule_match.id,
            rule_match.message,
        ]

    def output(self) -> str:
        return "\n".join(":".join(self._get_parts(rm)) for rm in self.rule_matches)



def test_envvar(run_semgrep_in_tmp):
    """
    Test metrics sending respects SEMGREP_SEND_METRICS envvar
    """
    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing", "SEMGREP_SEND_METRICS": "1"},
    )
    assert "Sent non-identifiable metrics" in output

    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing", "SEMGREP_SEND_METRICS": "true"},
    )
    assert "Sent non-identifiable metrics" in output

    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing", "SEMGREP_SEND_METRICS": ""},
    )
    assert "Sent non-identifiable metrics" not in output

    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing"},
    )
    assert "Sent non-identifiable metrics" not in output


def test_flags(run_semgrep_in_tmp):
    """
    Test metrics sending respects flags. Flags take precedence over envvar
    """
    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug", "--enable-metrics"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing"},
    )
    assert "Sent non-identifiable metrics" in output

    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug", "--enable-metrics"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing", "SEMGREP_SEND_METRICS": ""},
    )
    assert "Sent non-identifiable metrics" in output

    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug", "--disable-metrics"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing"},
    )
    assert "Sent non-identifiable metrics" not in output

    output = run_semgrep_in_tmp(
        "rules/eqeq.yaml",
        stderr=True,
        options=["--debug", "--disable-metrics"],
        env={"SEMGREP_USER_AGENT_APPEND": "testing", "SEMGREP_SEND_METRICS": "1"},
    )
    assert "Sent non-identifiable metrics" not in output



import contextlib
import json
import os
import subprocess
import sys
from pathlib import Path
from typing import List
from typing import Mapping
from typing import Optional
from typing import Union

import pytest

from semgrep.constants import OutputFormat

TESTS_PATH = Path(__file__).parent

MASKED_KEYS = [
    "tool.driver.semanticVersion",
    "results.extra.metavars.*.unique_id.md5sum",
    "results.*.checks.*.matches",
]


def mark_masked(obj, path):
    _mark_masked(obj, path.split("."))


def _mark_masked(obj, path_items):
    key = path_items[0]
    if len(path_items) == 1 and key in obj:
        obj[key] = "<masked in tests>"
    else:
        if key == "*":
            next_obj = list(obj.values())
        else:
            next_obj = obj.get(key)
        if next_obj is None:
            next_objs = []
        elif not isinstance(next_obj, list):
            next_objs = [next_obj]
        else:
            next_objs = next_obj
        for o in next_objs:
            if isinstance(o, dict):
                _mark_masked(o, path_items[1:])


def _clean_output_json(output_json: str) -> str:
    """Make semgrep's output deterministic and nicer to read."""
    output = json.loads(output_json)
    for path in MASKED_KEYS:
        mark_masked(output, path)

    return json.dumps(output, indent=2, sort_keys=True)


def _run_semgrep(
    config: Optional[Union[str, Path, List[str]]] = None,
    *,
    target_name: str = "basic",
    options: Optional[List[Union[str, Path]]] = None,
    output_format: OutputFormat = OutputFormat.JSON,
    stderr: bool = False,
    strict: bool = True,
    env: Optional[Mapping[str, str]] = None,
) -> str:
    """Run the semgrep CLI.

    :param config: what to pass as --config's value
    :param target_name: which directory within ./e2e/targets/ to scan
    :param options: additional CLI flags to add
    :param output_format: which format to use
    :param stderr: whether to merge stderr into the returned string
    """
    if options is None:
        options = []

    if strict:
        options.append("--strict")

    options.append("--disable-version-check")

    if config is not None:
        if isinstance(config, list):
            for conf in config:
                options.extend(["--config", conf])
        else:
            options.extend(["--config", config])

    if output_format == OutputFormat.JSON:
        options.append("--json")
    elif output_format == OutputFormat.JUNIT_XML:
        options.append("--junit-xml")
    elif output_format == OutputFormat.SARIF:
        options.append("--sarif")

    output = subprocess.check_output(
        [sys.executable, "-m", "semgrep", *options, Path("targets") / target_name],
        encoding="utf-8",
        stderr=subprocess.STDOUT if stderr else subprocess.PIPE,
        env=env,
    )

    if output_format.is_json() and not stderr:
        output = _clean_output_json(output)

    return output


@contextlib.contextmanager
def chdir(dirname=None):
    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)


@pytest.fixture
def run_semgrep_in_tmp(monkeypatch, tmp_path):
    (tmp_path / "targets").symlink_to(Path(TESTS_PATH / "e2e" / "targets").resolve())
    (tmp_path / "rules").symlink_to(Path(TESTS_PATH / "e2e" / "rules").resolve())

    monkeypatch.chdir(tmp_path)

    yield _run_semgrep



from pathlib import Path
from typing import Dict
from typing import NamedTuple
from typing import Optional

from semgrep.rule import Rule

Semgrep_run = NamedTuple("Semgrep_run", [("rule", Rule), ("target", Path)])

Times = NamedTuple(
    "Times", [("parse_time", float), ("match_time", float), ("run_time", float)]
)


class ProfilingData:
    def __init__(self) -> None:
        self._match_time_matrix: Dict[Semgrep_run, Times] = {}
        self._rule_parse_times: Dict[Rule, float] = {}
        self._file_parse_times: Dict[Path, float] = {}

        self._rule_match_times: Dict[Rule, float] = {}
        self._rule_run_times: Dict[Rule, float] = {}
        self._rule_bytes_scanned: Dict[Rule, int] = {}
        self._file_match_times: Dict[Path, float] = {}
        self._file_run_times: Dict[Path, float] = {}
        self._file_num_times_scanned: Dict[Path, int] = {}

    def get_run_times(self, rule: Rule, target: Path) -> Times:
        return self._match_time_matrix.get(
            Semgrep_run(rule=rule, target=target),
            Times(parse_time=0.0, match_time=0.0, run_time=0.0),
        )

    def get_file_parse_time(self, target: Path) -> Optional[float]:
        """
        Return time taken to parse a file. This is the max of all
        times reported to parse the file from semgrep-core since
        it caches the parsed file.

        Return None if target has no reported parse time
        """
        return self._file_parse_times.get(target)

    def get_rule_match_time(self, rule: Rule) -> Optional[float]:
        """
        Return total match time for a given rule over all the files scanned
        with said rule

        Return None if RULE has no timing information saved
        """
        return self._rule_match_times.get(rule)

    def get_rule_run_time(self, rule: Rule) -> Optional[float]:
        """
        Return total run time for a given rule over all the files scanned with
        said rule

        Return None if RULE has no timing information saved
        """
        return self._rule_run_times.get(rule)

    def get_rule_bytes_scanned(self, rule: Rule) -> int:
        """
        Return total number of bytes scanned by a given rule
        """
        return self._rule_bytes_scanned.get(rule, 0)

    def get_file_match_time(self, target: Path) -> Optional[float]:
        """
        Return total match time for a given file over all the rules that
        scanned the file

        Return None if TARGET has no timing information saved
        """
        return self._file_match_times.get(target)

    def get_file_run_time(self, target: Path) -> Optional[float]:
        """
        Return total run time for a given file over all the rules that
        scanned the file

        Return None if TARGET has no timing information saved
        """
        return self._file_run_times.get(target)

    def get_file_num_times_scanned(self, target: Path) -> int:
        """
        Returns number of times a file was scanned with rules.
        Assumes that each entry to set_run_times means a target
        was scanned once
        """
        return self._file_num_times_scanned.get(target, 0)

    def set_run_times(self, rule: Rule, target: Path, times: Times) -> None:
        num_bytes = target.stat().st_size

        self._rule_run_times[rule] = (
            self._rule_run_times.get(rule, 0.0) + times.run_time
        )
        self._rule_match_times[rule] = (
            self._rule_match_times.get(rule, 0.0) + times.match_time
        )
        self._rule_bytes_scanned[rule] = (
            self._rule_bytes_scanned.get(rule, 0) + num_bytes
        )

        self._file_num_times_scanned[target] = (
            self._file_num_times_scanned.get(target, 0) + 1
        )
        self._file_run_times[target] = (
            self._file_run_times.get(target, 0.0) + times.run_time
        )
        self._file_match_times[target] = (
            self._file_match_times.get(target, 0.0) + times.match_time
        )

        # File parse time is max of all parse times since others will be
        # cache hits
        self._file_parse_times[target] = max(
            times.parse_time, self._file_parse_times.get(target, 0.0)
        )

        self._match_time_matrix[Semgrep_run(rule=rule, target=target)] = times

    def get_rule_parse_time(self, rule: Rule) -> float:
        return self._rule_parse_times.get(rule, 0.0)

    def set_rule_parse_time(self, rule: Rule, parse_time: float) -> None:
        self._rule_parse_times[rule] = parse_time



import functools
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import NamedTuple
from typing import NewType
from typing import Optional

from semgrep.error import UnknownLanguageError
from semgrep.rule_lang import Span

Mode = NewType("Mode", str)
PatternId = NewType("PatternId", str)
FileExtension = NewType("FileExtension", str)

JOIN_MODE = Mode("join")
SEARCH_MODE = DEFAULT_MODE = Mode("search")


class Language(Enum):
    PYTHON: str = "python"
    PYTHON2: str = "python2"
    PYTHON3: str = "python3"
    JAVASCRIPT: str = "javascript"
    TYPESCRIPT: str = "typescript"
    JAVA: str = "java"
    C: str = "c"
    GO: str = "go"
    RUBY: str = "ruby"
    PHP: str = "php"
    LUA: str = "lua"
    CSHARP: str = "csharp"
    RUST: str = "rust"
    KOTLIN: str = "kt"
    YAML: str = "yaml"
    ML: str = "ml"
    SCALA: str = "scala"
    VUE: str = "vue"
    HTML: str = "html"
    JSON: str = "json"
    REGEX: str = "regex"
    GENERIC: str = "generic"


class Language_util:
    language_to_strs: Dict[Language, List[str]] = {
        Language.PYTHON: [Language.PYTHON.value, "py"],
        Language.PYTHON2: [Language.PYTHON2.value],
        Language.PYTHON3: [Language.PYTHON3.value],
        Language.JAVASCRIPT: [Language.JAVASCRIPT.value, "js"],
        Language.TYPESCRIPT: [Language.TYPESCRIPT.value, "ts"],
        Language.JAVA: [Language.JAVA.value],
        Language.C: [Language.C.value],
        Language.GO: [Language.GO.value, "golang"],
        Language.RUBY: [Language.RUBY.value, "rb"],
        Language.PHP: [Language.PHP.value],
        Language.LUA: [Language.LUA.value],
        Language.CSHARP: [Language.CSHARP.value, "cs", "C#"],
        Language.RUST: [Language.RUST.value, "Rust", "rs"],
        Language.KOTLIN: [Language.KOTLIN.value, "Kotlin", "kotlin"],
        Language.YAML: [Language.YAML.value, "Yaml"],
        Language.SCALA: [Language.SCALA.value],
        Language.VUE: [Language.VUE.value],
        Language.HTML: [Language.HTML.value],
        Language.ML: [Language.ML.value, "ocaml"],
        Language.JSON: [Language.JSON.value, "JSON", "Json"],
        Language.REGEX: [Language.REGEX.value, "none"],
        Language.GENERIC: [Language.GENERIC.value],
    }

    str_to_language = {}
    for language in language_to_strs.keys():
        for lang_str in language_to_strs[language]:
            str_to_language[lang_str] = language

    """ Convert an inputted string representing a language to a Language

    Keyword arguments;
    lang_str -- string representing a language (e.g. "C#")
    span     -- span of language string in the original file (for error reporting),
                None if resolve was called within semgrep
    """

    @classmethod
    def resolve(cls, lang_str: str, span: Optional[Span] = None) -> Language:
        if lang_str in cls.str_to_language:
            return cls.str_to_language[lang_str]
        else:
            raise UnknownLanguageError(
                short_msg=f"invalid language: {lang_str}",
                long_msg=f"unsupported language: {lang_str}. supported languages are: {', '.join(cls.all_language_strs())}",
                spans=[span.with_context(before=1, after=1)]
                if span
                else [],  # not called from a config file
            )

    @classmethod
    def all_language_strs(cls) -> List[str]:
        # sort to standardize because this is used in outputting methods
        return sorted(cls.str_to_language.keys())


ALLOWED_GLOB_TYPES = ("include", "exclude")


@functools.total_ordering
class Range(NamedTuple):
    start: int
    end: int
    metavariables: Mapping[str, Any]
    propagated_metavariables: Dict[str, Any] = {}

    def metavariables_match(self, rhs: "Range") -> bool:
        """
        Returns true if and only if all metavariables in both this and the other Range
        refer to the same variables (if variable nodes), in the same scope, or the same
        expressions (if expression nodes).

        That is, if two patterns define a "$X", and $X refers to a variable in one
        pattern, then $X must refer to the same variable in both patterns.
        """
        to_match = set(self.metavariables.keys()) & set(rhs.metavariables.keys())
        return all(
            self.metavariables[v]
            and rhs.metavariables[v]
            and self.metavariables[v] == rhs.metavariables[v]
            for v in to_match
        )

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} "
            f"start={self.start} "
            f"end={self.end} "
            f"metavariables={self.metavariables} "
            f"propagated={self.propagated_metavariables}>"
        )

    def __hash__(self) -> int:
        return hash((self.start, self.end))

    def __eq__(self, rhs: Any) -> bool:
        if not isinstance(rhs, type(self)):
            return False

        return (
            self.start == rhs.start
            and self.end == rhs.end
            and self.metavariables_match(rhs)
        )

    def __lt__(self, rhs: Any) -> bool:
        if not isinstance(rhs, type(self)):
            return False

        diff_self = self.end - self.start
        diff_rhs = rhs.end - rhs.start

        if diff_self == diff_rhs:
            return self.start < rhs.start

        return diff_self < diff_rhs



from typing import List

from semgrep.constants import CLI_RULE_ID
from semgrep.formatter.base import BaseFormatter
from semgrep.rule_match import RuleMatch


class EmacsFormatter(BaseFormatter):
    @staticmethod
    def _get_parts(rule_match: RuleMatch) -> List[str]:
        check_id = (
            rule_match.id.split(".")[-1] if rule_match.id != CLI_RULE_ID else None
        )
        severity = (
            rule_match.severity.lower() + f"({check_id})"
            if check_id
            else rule_match.severity.lower()
        )
        return [
            str(rule_match.path),
            str(rule_match.start["line"]),
            str(rule_match.start["col"]),
            severity,
            rule_match.lines[0].rstrip(),
        ]

    def output(self) -> str:
        rule_matches = sorted(self.rule_matches, key=lambda r: (r.path, r.id))
        return "\n".join(":".join(self._get_parts(rm)) for rm in rule_matches)



import hashlib
import json
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from semgrep.error import InvalidRuleSchemaError
from semgrep.rule_lang import EmptySpan
from semgrep.rule_lang import Span
from semgrep.rule_lang import YamlMap
from semgrep.rule_lang import YamlTree
from semgrep.semgrep_types import ALLOWED_GLOB_TYPES
from semgrep.semgrep_types import JOIN_MODE
from semgrep.semgrep_types import Language
from semgrep.semgrep_types import Language_util
from semgrep.semgrep_types import SEARCH_MODE


class Rule:
    def __init__(self, raw: YamlTree[YamlMap]) -> None:
        self._yaml = raw
        self._raw: Dict[str, Any] = raw.unroll_dict()

        self._id = str(self._raw["id"])

        paths_tree: Optional[YamlTree] = self._yaml.value.get("paths")
        if paths_tree is None:
            path_dict = {}
        else:
            paths, paths_span = paths_tree.value, paths_tree.span
            if not isinstance(paths, YamlMap):
                path_key = self._yaml.value.key_tree("paths").span
                help_str: Optional[str] = None
                if isinstance(paths, list):
                    help_str = "remove the `-` to convert the list into a mapping"
                raise InvalidRuleSchemaError(
                    short_msg="invalid paths",
                    long_msg=f"the `paths:` targeting rules must be an object with at least one of {ALLOWED_GLOB_TYPES}",
                    spans=[path_key.extend_to(paths_span)],
                    help=help_str,
                )
            path_dict = paths_tree.unroll_dict()
        self._includes = path_dict.get("include", [])
        self._excludes = path_dict.get("exclude", [])
        rule_languages = {
            Language_util.resolve(l, self.languages_span)
            for l in self._raw.get("languages", [])
        }

        # add typescript to languages if the rule supports javascript.
        if any(language == Language.JAVASCRIPT for language in rule_languages):
            rule_languages.add(Language.TYPESCRIPT)
            self._raw["languages"] = [r.value for r in rule_languages]

        self._languages = sorted(rule_languages, key=lambda lang: lang.value)  # type: ignore

        # check taint/search mode
        if self._raw.get("mode") == JOIN_MODE:
            self._mode = JOIN_MODE
        else:
            self._mode = SEARCH_MODE

        if any(language == Language.REGEX for language in self._languages):
            self._validate_none_language_rule()

    def _validate_none_language_rule(self) -> None:
        """
        For regex-only rules, only patterns, pattern-either, and pattern-regex is valid.
        """

        def _recursive_contains(
            obj: Union[Dict[str, Any], List[Any], str], search_key: str
        ) -> bool:
            """
            Returns true if object contains any object that contains search_key as key
            """
            if isinstance(obj, dict):
                for key in obj:
                    if key == search_key:
                        return True

                    if _recursive_contains(obj[key], search_key):
                        return True

            if isinstance(obj, list):
                for elem in obj:
                    if _recursive_contains(elem, search_key):
                        return True

            return False

        if _recursive_contains(self._raw, "pattern"):
            raise InvalidRuleSchemaError(
                short_msg=f"invalid pattern clause",
                long_msg=f"invalid pattern clause 'pattern' with regex-only rules in rule: {self.id}",
                spans=[],
                help=f"use only patterns, pattern-either, pattern-regex, or pattern-not-regex with regex-only rules",
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False

        return self._raw == other._raw

    def __hash__(self) -> int:
        return hash(self.id)

    @property
    def includes(self) -> List[str]:
        return self._includes  # type: ignore

    @property
    def excludes(self) -> List[str]:
        return self._excludes  # type: ignore

    @property
    def id(self) -> str:
        return self._id

    @property
    def message(self) -> str:
        return str(self._raw["message"])

    @property
    def metadata(self) -> Dict[str, Any]:  # type: ignore
        return self._raw.get("metadata", {})

    @property
    def severity(self) -> str:
        return str(self._raw["severity"])

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def languages(self) -> List[Language]:
        return self._languages

    @property
    def languages_span(self) -> Span:
        return self._yaml.value["languages"].span

    @property
    def raw(self) -> Dict[str, Any]:  # type: ignore
        return self._raw

    @property
    def fix(self) -> Optional[str]:
        return self._raw.get("fix")

    @property
    def fix_regex(self) -> Optional[Dict[str, Any]]:  # type: ignore
        return self._raw.get("fix-regex")

    @classmethod
    def from_json(cls, rule_json: Dict[str, Any]) -> "Rule":  # type: ignore
        yaml = YamlTree.wrap(rule_json, EmptySpan)
        return cls(yaml)

    @classmethod
    def from_yamltree(cls, rule_yaml: YamlTree[YamlMap]) -> "Rule":
        return cls(rule_yaml)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"

    def rename_id(self, new_id: str) -> None:
        self._id = new_id

    @property
    def full_hash(self) -> str:
        """
        sha256 hash of the whole rule object instead of just the id
        """
        return hashlib.sha256(
            json.dumps(self._raw, sort_keys=True).encode()
        ).hexdigest()



import copy
import json
from typing import Any
from typing import Dict

from semgrep.formatter.base import BaseFormatter
from semgrep.rule_match import RuleMatch


class JsonFormatter(BaseFormatter):
    @staticmethod
    def _rule_match_to_json(rule_match: RuleMatch) -> Dict[str, Any]:
        json_obj = copy.deepcopy(rule_match._pattern_match._raw_json)

        json_obj["check_id"] = rule_match.id
        json_obj["extra"] = json_obj.get("extra", {})
        json_obj["extra"]["message"] = rule_match.message
        json_obj["extra"]["metadata"] = rule_match.metadata
        json_obj["extra"]["severity"] = rule_match.severity
        json_obj["path"] = json_obj.get("path", str(rule_match.path))
        json_obj["start"] = rule_match.start
        json_obj["end"] = rule_match.end

        # 'lines' already contains '\n' at the end of each line
        json_obj["extra"]["lines"] = "".join(rule_match.lines).rstrip()

        if rule_match.fix:
            json_obj["extra"]["fix"] = rule_match.fix
        if rule_match.fix_regex:
            json_obj["extra"]["fix_regex"] = rule_match.fix_regex
        if rule_match.is_ignored is not None:
            json_obj["extra"]["is_ignored"] = rule_match.is_ignored

        return json_obj

    def output(self) -> str:
        output_dict = {
            "results": [
                self._rule_match_to_json(rule_match) for rule_match in self.rule_matches
            ],
            "errors": [error.to_dict() for error in self.semgrep_structured_errors],
            **self.extra,
        }
        # Sort keys for predictable output. This helps with snapshot tests, etc.
        return json.dumps(output_dict, sort_keys=True)



from typing import Dict
from typing import List
from typing import Optional
from typing import Set

from semgrep.pattern_match import PatternMatch
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.semgrep_types import Range
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


def compare_propagated_metavariable(
    _range: Range,
    pattern_match: PatternMatch,
    metavariable: str,
) -> bool:

    return (
        metavariable in _range.propagated_metavariables
        and metavariable in pattern_match.metavariable_uids
        and _range.propagated_metavariables[metavariable]
        == pattern_match.metavariable_uids[metavariable]
    )


def create_output(
    rule: Rule,
    pattern_matches: List[PatternMatch],
    valid_ranges_to_output: Optional[Set[Range]] = None,
) -> List[RuleMatch]:
    output = []

    if valid_ranges_to_output is None:
        valid_ranges_to_output = {
            pattern_match.range for pattern_match in pattern_matches
        }

    propagated_metavariable_lookup = {
        _range: {
            metavariable: pm.get_metavariable_value(metavariable)
            for pm in pattern_matches
            for metavariable in _range.propagated_metavariables
            if compare_propagated_metavariable(_range, pm, metavariable)
        }
        for _range in valid_ranges_to_output
    }

    for pattern_match in pattern_matches:
        if pattern_match.range in valid_ranges_to_output:
            propagated_metavariables = propagated_metavariable_lookup[
                pattern_match.range
            ]
            message = interpolate_string_with_metavariables(
                rule.message, pattern_match, propagated_metavariables
            )
            fix = (
                interpolate_string_with_metavariables(
                    rule.fix, pattern_match, propagated_metavariables
                )
                if rule.fix
                else None
            )
            rule_match = RuleMatch.from_pattern_match(
                rule.id,
                pattern_match,
                message=message,
                metadata=rule.metadata,
                severity=rule.severity,
                fix=fix,
                fix_regex=rule.fix_regex,
            )
            output.append(rule_match)

    return sorted(output, key=lambda rule_match: rule_match._pattern_match.range.start)


def interpolate_string_with_metavariables(
    text: str, pattern_match: PatternMatch, propagated_metavariables: Dict[str, str]
) -> str:
    """Interpolates a string with the metavariables contained in it, returning a new string"""

    # Sort by metavariable length to avoid name collisions (eg. $X2 must be handled before $X)
    for metavariable in sorted(pattern_match.metavariables, key=len, reverse=True):
        text = text.replace(
            metavariable, pattern_match.get_metavariable_value(metavariable)
        )

    # Sort by metavariable length to avoid name collisions (eg. $X2 must be handled before $X)
    for metavariable in sorted(propagated_metavariables.keys(), key=len, reverse=True):
        text = text.replace(metavariable, propagated_metavariables[metavariable])

    return text



from typing import cast

from semgrep.external.junit_xml import TestCase  # type: ignore[attr-defined]
from semgrep.external.junit_xml import TestSuite  # type: ignore[attr-defined]
from semgrep.external.junit_xml import to_xml_report_string  # type: ignore[attr-defined]
from semgrep.formatter.base import BaseFormatter
from semgrep.rule_match import RuleMatch


class JunitXmlFormatter(BaseFormatter):
    @staticmethod
    def _rule_match_to_test_case(rule_match: RuleMatch) -> TestCase:  # type: ignore
        test_case = TestCase(
            rule_match.id,
            file=str(rule_match.path),
            line=rule_match.start["line"],
            classname=str(rule_match.path),
        )
        test_case.add_failure_info(
            message=rule_match.message,
            output=rule_match.lines,
            failure_type=rule_match.severity,
        )
        return test_case

    def output(self) -> str:
        test_cases = [
            self._rule_match_to_test_case(rule_match)
            for rule_match in self.rule_matches
        ]
        ts = TestSuite("semgrep results", test_cases)
        return cast(str, to_xml_report_string([ts]))



import collections
import json
import subprocess
import tempfile
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

from ruamel.yaml import YAML

from semgrep.constants import PLEASE_FILE_ISSUE_TEXT
from semgrep.core_exception import CoreException
from semgrep.error import _UnknownLanguageError
from semgrep.error import InvalidPatternError
from semgrep.error import MatchTimeoutError
from semgrep.error import SemgrepError
from semgrep.error import UnknownLanguageError
from semgrep.evaluation import create_output
from semgrep.pattern_match import PatternMatch
from semgrep.profile_manager import ProfileManager
from semgrep.profiling import ProfilingData
from semgrep.profiling import Times
from semgrep.rule import Rule
from semgrep.rule_lang import Span
from semgrep.rule_match import RuleMatch
from semgrep.semgrep_types import Language
from semgrep.target_manager import TargetManager
from semgrep.target_manager_extensions import all_supported_languages
from semgrep.util import debug_tqdm_write
from semgrep.util import is_debug
from semgrep.util import progress_bar
from semgrep.util import SEMGREP_PATH
from semgrep.util import sub_run
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


class CoreRunner:
    """
    Handles interactions between semgrep and semgrep-core

    This includes properly invoking semgrep-core and parsing the output
    """

    def __init__(
        self,
        jobs: int,
        timeout: int,
        max_memory: int,
        timeout_threshold: int,
        optimizations: str,
    ):
        self._jobs = jobs
        self._timeout = timeout
        self._max_memory = max_memory
        self._timeout_threshold = timeout_threshold
        self._optimizations = optimizations

    def _raise_semgrep_error_from_json(
        self,
        error_json: Dict[str, Any],
        rule: Rule,
    ) -> None:
        """
        See format_output_exception in semgrep O'Caml for details on schema
        """
        error_type = error_json["error"]
        if error_type == "invalid language":
            raise SemgrepError(
                f'{error_json["language"]} was accepted by semgrep but rejected by semgrep-core. {PLEASE_FILE_ISSUE_TEXT}'
            )
        elif error_type == "invalid regexp in rule":
            raise SemgrepError(f'Invalid regexp in rule: {error_json["message"]}')
        elif error_type == "invalid pattern":
            range = error_json["range"]
            s = error_json.get("pattern", "<no pattern>")
            matching_span = Span.from_string_token(
                s=s,
                line=range.get("line", 0),
                col=range.get("col", 0),
                path=range.get("path", []),
                filename="semgrep temp file",
            )
            if error_json["message"] == "Parsing.Parse_error":
                long_msg = f"Pattern `{s.strip()}` could not be parsed as a {error_json['language']} semgrep pattern"
            else:
                long_msg = f"Error parsing {error_json['language']} pattern: {error_json['message']}"

            raise InvalidPatternError(
                short_msg=error_type,
                long_msg=long_msg,
                spans=[matching_span],
                help=None,
            )
        # no special formatting ought to be required for the other types; the semgrep python should be performing
        # validation for them. So if any other type of error occurs, ask the user to file an issue
        else:
            raise SemgrepError(
                f"an internal error occured while invoking semgrep-core while running rule '{rule.id}'. Consider skipping this rule and reporting this issue.\n\t{error_type}: {error_json.get('message', 'no message')}\n{PLEASE_FILE_ISSUE_TEXT}"
            )

    def _extract_core_output(
        self, rule: Rule, core_run: subprocess.CompletedProcess
    ) -> Dict[str, Any]:
        semgrep_output = core_run.stdout.decode("utf-8", errors="replace")

        stderr = core_run.stderr
        if stderr is None:
            semgrep_error_output = (
                "<semgrep-core stderr not captured, should be printed above>\n"
            )
        else:
            semgrep_error_output = stderr.decode("utf-8", errors="replace")

        # By default, we print semgrep-core's error output, which includes
        # semgrep-core's logging if it was requested via --debug.
        #
        # If semgrep-core prints anything on stderr when running with default
        # flags, it's a bug that should be fixed in semgrep-core.
        #
        name = f"[rule '{rule.id}']"
        logger.debug(
            f"--- semgrep-core stderr {name} ---\n"
            f"{semgrep_error_output}"
            f"--- end semgrep-core stderr {name} ---"
        )

        returncode = core_run.returncode
        if returncode != 0:
            output_json = self._parse_core_output(
                rule, core_run, semgrep_output, semgrep_error_output, returncode
            )

            if "error" in output_json:
                self._raise_semgrep_error_from_json(output_json, rule)
            else:
                self._fail(
                    'non-zero exit status with missing "error" field in json response',
                    rule,
                    core_run,
                    returncode,
                    semgrep_output,
                    semgrep_error_output,
                )

        output_json = self._parse_core_output(
            rule, core_run, semgrep_output, semgrep_error_output, returncode
        )
        return output_json

    def _parse_core_output(
        self,
        rule: Rule,
        core_run: subprocess.CompletedProcess,
        semgrep_output: str,
        semgrep_error_output: str,
        returncode: int,
    ) -> Dict[str, Any]:
        # See if semgrep output contains a JSON error that we can decode.
        try:
            return cast(Dict[str, Any], json.loads(semgrep_output))
        except ValueError:
            self._fail(
                "unparseable json output",
                rule,
                core_run,
                returncode,
                semgrep_output,
                semgrep_error_output,
            )
            return {}  # never reached

    def _fail(
        self,
        reason: str,
        rule: Rule,
        core_run: subprocess.CompletedProcess,
        returncode: int,
        semgrep_output: str,
        semgrep_error_output: str,
    ) -> None:
        # Once we require python >= 3.8, switch to using shlex.join instead
        # for proper quoting of the command line.
        shell_command = " ".join(core_run.args)
        raise SemgrepError(
            f"semgrep-core failed: {reason}\n"
            f"rule ID: '{rule.id}'\n"
            f"semgrep-core exit code: {returncode}\n"
            f"semgrep-core command: {shell_command}\n"
            f"unexpected non-json output while invoking semgrep-core:\n"
            "--- semgrep-core stdout ---\n"
            f"{semgrep_output}"
            "--- end semgrep-core stdout ---\n"
            "--- semgrep-core stderr ---\n"
            f"{semgrep_error_output}"
            "--- end semgrep-core stderr ---\n"
            f"{PLEASE_FILE_ISSUE_TEXT}"
        )

    def _add_match_times(
        self,
        rule: Rule,
        profiling_data: ProfilingData,
        output_time_json: Dict[str, Any],
    ) -> None:
        """Collect the match times reported by semgrep-core (or spacegrep)."""
        if "targets" in output_time_json:
            for target in output_time_json["targets"]:
                if "match_time" in target and "path" in target:
                    profiling_data.set_run_times(
                        rule,
                        Path(target["path"]),
                        Times(
                            parse_time=target["parse_time"],
                            match_time=target["match_time"],
                            run_time=target["run_time"],
                        ),
                    )
        if "rule_parse_time" in output_time_json:
            profiling_data.set_rule_parse_time(
                rule, output_time_json["rule_parse_time"]
            )

    @staticmethod
    def get_files_for_language(
        language: Language, rule: Rule, target_manager: TargetManager
    ) -> List[Path]:
        try:
            targets = target_manager.get_files(language, rule.includes, rule.excludes)
        except _UnknownLanguageError as ex:
            raise UnknownLanguageError(
                short_msg=f"invalid language: {language}",
                long_msg=f"unsupported language: {language}. supported languages are: {', '.join(all_supported_languages())}",
                spans=[rule.languages_span.with_context(before=1, after=1)],
            ) from ex
        return list(targets)

    def _run_rules_direct_to_semgrep_core(
        self,
        rules: List[Rule],
        target_manager: TargetManager,
        profiler: ProfileManager,
    ) -> Tuple[
        Dict[Rule, List[RuleMatch]],
        Dict[Rule, List[Any]],
        List[SemgrepError],
        Set[Path],
        ProfilingData,
    ]:
        logger.debug(f"Passing whole rules directly to semgrep_core")

        outputs: Dict[Rule, List[RuleMatch]] = collections.defaultdict(list)
        errors: List[SemgrepError] = []
        all_targets: Set[Path] = set()
        file_timeouts: Dict[Path, int] = collections.defaultdict(lambda: 0)
        max_timeout_files: Set[Path] = set()

        profiling_data: ProfilingData = ProfilingData()
        # cf. for bar_format: https://tqdm.github.io/docs/tqdm/
        with tempfile.TemporaryDirectory() as semgrep_core_ast_cache_dir:
            for rule in progress_bar(
                rules, bar_format="{l_bar}{bar}|{n_fmt}/{total_fmt}"
            ):
                for language in rule.languages:
                    debug_tqdm_write(f"Running rule {rule.id}...")
                    with tempfile.NamedTemporaryFile(
                        "w", suffix=".yaml"
                    ) as rule_file, tempfile.NamedTemporaryFile(
                        "w"
                    ) as target_file, tempfile.NamedTemporaryFile(
                        "w"
                    ) as equiv_file:
                        targets = self.get_files_for_language(
                            language, rule, target_manager
                        )

                        targets = [
                            target
                            for target in targets
                            if target not in max_timeout_files
                        ]

                        # opti: no need to call semgrep-core if no target files
                        if not targets:
                            continue
                        all_targets = all_targets.union(targets)

                        target_file.write("\n".join(map(lambda p: str(p), targets)))
                        target_file.flush()
                        yaml = YAML()
                        yaml.dump({"rules": [rule._raw]}, rule_file)
                        rule_file.flush()

                        cmd = [SEMGREP_PATH] + [
                            "-lang",
                            language.value,
                            "-json",
                            "-config",
                            rule_file.name,
                            "-j",
                            str(self._jobs),
                            "-target_file",
                            target_file.name,
                            "-use_parsing_cache",
                            semgrep_core_ast_cache_dir,
                            "-timeout",
                            str(self._timeout),
                            "-max_memory",
                            str(self._max_memory),
                            "-json_time",
                        ]

                        if self._optimizations != "none":
                            cmd.append("-fast")

                        stderr: Optional[int] = subprocess.PIPE
                        if is_debug():
                            cmd += ["-debug"]
                            stderr = None

                        core_run = sub_run(cmd, stdout=subprocess.PIPE, stderr=stderr)
                        output_json = self._extract_core_output(rule, core_run)

                        if "time" in output_json:
                            self._add_match_times(
                                rule, profiling_data, output_json["time"]
                            )

                    # end with tempfile.NamedTemporaryFile(...) ...
                    pattern_matches = [
                        PatternMatch(match) for match in output_json["matches"]
                    ]
                    findings = create_output(rule, pattern_matches)

                    findings = dedup_output(findings)
                    outputs[rule].extend(findings)
                    parsed_errors = [
                        CoreException.from_json(
                            e, language.value, rule.id
                        ).into_semgrep_error()
                        for e in output_json["errors"]
                    ]
                    for err in parsed_errors:
                        if isinstance(err, MatchTimeoutError):
                            file_timeouts[err.path] += 1
                            if (
                                self._timeout_threshold != 0
                                and file_timeouts[err.path] >= self._timeout_threshold
                            ):
                                max_timeout_files.add(err.path)
                    errors.extend(parsed_errors)
            # end for language ...
        # end for rule ...

        return outputs, {}, errors, all_targets, profiling_data

    # end _run_rules_direct_to_semgrep_core

    def invoke_semgrep(
        self,
        target_manager: TargetManager,
        profiler: ProfileManager,
        rules: List[Rule],
    ) -> Tuple[
        Dict[Rule, List[RuleMatch]],
        Dict[Rule, List[Dict[str, Any]]],
        List[SemgrepError],
        Set[Path],
        ProfilingData,
    ]:
        """
        Takes in rules and targets and retuns object with findings
        """
        start = datetime.now()

        (
            findings_by_rule,
            debug_steps_by_rule,
            errors,
            all_targets,
            profiling_data,
        ) = self._run_rules_direct_to_semgrep_core(rules, target_manager, profiler)

        logger.debug(
            f"semgrep ran in {datetime.now() - start} on {len(all_targets)} files"
        )
        by_severity = collections.defaultdict(list)
        for rule, findings in findings_by_rule.items():
            by_severity[rule.severity.lower()].extend(findings)

        by_sev_strings = [
            f"{len(findings)} {sev}" for sev, findings in by_severity.items()
        ]
        logger.debug(f'findings summary: {", ".join(by_sev_strings)}')

        return (
            findings_by_rule,
            debug_steps_by_rule,
            errors,
            all_targets,
            profiling_data,
        )


# This will remove matches that have the same range but different
# metavariable bindings, choosing the last one in the list. We want the
# last because if there multiple possible bindings, they will be returned
# by semgrep-core from largest range to smallest. For an example, see
# tests/e2e/test_message_interpolation.py::test_message_interpolation;
# specifically, the multi-pattern-inside test
#
# Another option is to not dedup, since Semgrep.ml now does its own deduping
# otherwise, and surface both matches
def dedup_output(outputs: List[RuleMatch]) -> List[RuleMatch]:
    return list({uniq_id(r): r for r in reversed(outputs)}.values())[::-1]


def uniq_id(
    r: RuleMatch,
) -> Tuple[str, Path, Optional[int], Optional[int], Optional[int], Optional[int], str]:
    start = r.start
    end = r.end
    return (
        r.id,
        r.path,
        start.get("line"),
        start.get("col"),
        end.get("line"),
        end.get("col"),
        r.message,
    )



#
# Parse command-line arguments representing a number of bytes, such as
# '5 mb' or '3.2GiB'.
#
# Maybe we could use an external package for this, such as
# https://pypi.org/project/datasize/ (documentation?)
# or https://github.com/xolox/python-humanfriendly
#
import re

UNITS = {
    "": 1,
    "B": 1,
    "KIB": 2 ** 10,
    "MIB": 2 ** 20,
    "GIB": 2 ** 30,
    "TIB": 2 ** 40,
    "KB": 10 ** 3,
    "MB": 10 ** 6,
    "GB": 10 ** 9,
    "TB": 10 ** 12,
}


def parse_size(input: str) -> int:
    s = input.upper()
    # note that '1e6' is a valid float and should not become '1 e6'.
    s = re.sub(r"([BKMGT][A-Z]*)", r" \1", s)
    tokens = [sub.strip() for sub in s.split()]
    n = len(tokens)
    if n == 1:
        number = tokens[0]
        unit = ""
    elif n == 2:
        number, unit = tokens
    else:
        raise ValueError(f"Invalid representation for a number of bytes: '{input}'")
    if unit in UNITS:
        return int(float(number) * UNITS[unit])
    else:
        raise ValueError(f"Invalid representation for a number of bytes: '{input}'")



import itertools
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import attr

from semgrep.pattern_match import PatternMatch


@attr.s(frozen=True)
class RuleMatch:
    """
    A section of code that matches a single rule (which is potentially many patterns)
    """

    _id: str = attr.ib()
    _pattern_match: PatternMatch = attr.ib(repr=False)
    _message: str = attr.ib(repr=False)
    _metadata: Dict[str, Any] = attr.ib(repr=False)
    _severity: str = attr.ib(repr=False)
    _fix: Optional[str] = attr.ib(repr=False)
    _fix_regex: Optional[Dict[str, Any]] = attr.ib(repr=False)

    # derived attributes
    _path: Path = attr.ib(repr=str)
    _start: Dict[str, Any] = attr.ib(repr=str)
    _end: Dict[str, Any] = attr.ib(repr=str)
    _extra: Dict[str, Any] = attr.ib(repr=False)
    _lines_cache: Dict[Tuple[str, str], List[str]] = attr.ib(repr=False)

    # optional attributes
    _is_ignored: Optional[bool] = attr.ib(default=None)

    @classmethod
    def from_pattern_match(
        cls,
        rule_id: str,
        pattern_match: PatternMatch,
        message: str,
        metadata: Dict[str, Any],
        severity: str,
        fix: Optional[str],
        fix_regex: Optional[Dict[str, Any]],
    ) -> "RuleMatch":
        path = pattern_match.path
        start = pattern_match.start
        end = pattern_match.end

        # note that message in extra is still the old value defined before metavar interpolation
        extra = pattern_match.extra
        return cls(
            rule_id,
            pattern_match,
            message,
            metadata,
            severity,
            fix,
            fix_regex,
            path,
            start,
            end,
            extra,
            {},
        )

    @property
    def id(self) -> str:
        return self._id

    @property
    def path(self) -> Path:
        return self._path

    @property
    def extra(self) -> Dict[str, Any]:
        return self._extra

    @property
    def fix(self) -> Optional[str]:
        # We do this for consistency with semgrep-core, which ignores whitespace
        # including newline chars at the end of multiline patterns
        return self._fix.rstrip() if self._fix else self._fix

    @property
    def fix_regex(self) -> Optional[Dict[str, Any]]:
        return self._fix_regex

    @property
    def message(self) -> str:
        return self._message

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @property
    def severity(self) -> str:
        return self._severity

    @property
    def start(self) -> Dict[str, Any]:
        return self._start

    @property
    def end(self) -> Dict[str, Any]:
        return self._end

    @property
    def is_ignored(self) -> Optional[bool]:
        return self._is_ignored

    @property
    def lines(self) -> List[str]:
        """
        Return lines in file that this RuleMatch is referring to.

        Assumes file exists.
        """
        # Start and end line are one-indexed, but the subsequent slice call is
        # inclusive for start and exclusive for end, so only subtract from start
        start_line = self.start["line"] - 1
        end_line = self.end["line"]

        try:
            return self._lines_cache[(start_line, end_line)]
        except KeyError:
            pass

        # buffering=1 turns on line-level reads
        with self.path.open(buffering=1, errors="replace") as fd:
            result = list(itertools.islice(fd, start_line, end_line))

        self._lines_cache[(start_line, end_line)] = result
        return result
